#
# Copyright (C) 2006,2013  Kipp Cannon
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


#
# =============================================================================
#
#                                   Preamble
#
# =============================================================================
#


from optparse import OptionParser


from glue import git_version
from glue.ligolw import ligolw
from glue.ligolw import lsctables
from glue.ligolw import utils as ligolw_utils
from glue.ligolw.utils import process as ligolw_process
from glue.ligolw.utils import search_summary as ligolw_search_summary
from glue.ligolw.utils import segments as ligolw_segments


class LIGOLWContentHandler(ligolw.LIGOLWContentHandler):
	pass
lsctables.use_in(LIGOLWContentHandler)


__author__ = "Kipp Cannon <kipp.cannon@ligo.org>"
__version__ = "git id %s" % git_version.id
__date__ = git_version.date


#
# =============================================================================
#
#                                 Command Line
#
# =============================================================================
#


parser = OptionParser(
	version = "Name: %%prog\n%s" % git_version.verbose_msg
)
parser.add_option("-c", "--comment", metavar = "text", default = "", help = "Set process comment.")
parser.add_option("-i", "--input", metavar = "filename", help = "Read from filename (default = stdin).")
parser.add_option("-o", "--output", metavar = "filename", help = "Write to filename (default = stdout).")
parser.add_option("-v", "--verbose", action = "store_true", help = "Be verbose.")
options, args = parser.parse_args()
del parser, args


#
# =============================================================================
#
#                                    Input
#
# =============================================================================
#


indoc = ligolw_utils.load_filename(options.input, contenthandler = LIGOLWContentHandler, verbose = options.verbose)


#
# =============================================================================
#
#                              Output Preparation
#
# =============================================================================
#


doc = ligolw.Document()
doc.appendChild(ligolw.LIGO_LW())
process = ligolw_process.register_to_xmldoc(doc, "ligolw_inspiral2mon", options.__dict__, version = __version, cvs_repository = "lscsoft", cvs_entry_time = __date__, comment = options.comment)
ligolwmontable = lsctables.New(lsctables.LIGOLWMonTable, columns = ["process_id", "time", "time_ns", "amplitude", "confidence", "frequency", "event_id"])
doc.childNodes[0].appendChild(ligolwmontable)


#
# =============================================================================
#
#                             Build Segment Table
#
# =============================================================================
#


seglists = lsctables.SearchSummaryTable.get_table(indoc).get_out_segmentlistdict(lsctables.ProcessTable.get_table(indoc).get_ids_by_program("inspiral")).coalesce()
segs = ligolw_segments.LigolwSegments(doc)
segs.insert_from_segmentlistdict(seglists, "inspiral", comment = options.comment)
segs.optimize()
segs.finalize(process)
ligolw_search_summary.append_search_summary(doc, process, comment = options.comment, inseg = seglists.extent_all(), outseg = seglists.extent_all())


#
# =============================================================================
#
#                     Convert sngl_inspiral To ligolw_mon
#
# =============================================================================
#


for inspiral in lsctables.SnglInspiralTable.get_table(indoc):
	ligolwmon = lsctables.LIGOLWMon()
	ligolwmon.process_id = process.process_id
	ligolwmon.event_id = ligolwmontable.get_next_id()
	ligolwmon.set_time(inspiral.get_end())
	ligolwmon.amplitude = inspiral.snr
	ligolwmon.confidence = inspiral.chisq
	ligolwmon.frequency = inspiral.f_final
	ligolwmontable.append(ligolwmon)


#
# =============================================================================
#
#                                    Output
#
# =============================================================================
#


ligolw_process.set_process_end_time(process)


ligolw_utils.write_filename(doc, options.output, gz = (options.output or "stdout").endswith(".gz"), verbose = options.verbose)



#
#
# Copyright (C) 2009  Larne Pekowsky
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


#
# =============================================================================
#
#                                   Preamble
#
# =============================================================================
#


"""
Takes one or more segment XML files and a query file, runs the query against
the segments, and produces veto files as XML
"""


from __future__ import print_function
from optparse import OptionParser
import sqlite3
import sys
import os
import pwd
import tempfile
import operator
import ligo.segments

import time
from glue import gpstime

from glue.ligolw import ligolw
from glue.ligolw import lsctables
from glue.ligolw import utils
from glue.ligolw.utils import ligolw_sqlite
from glue.ligolw import dbtables

from glue.segmentdb import query_engine
from glue.segmentdb import segmentdb_utils

from glue.ligolw.utils import process

from glue import git_version

import re
import urllib

PROGRAM_NAME = sys.argv[0].replace('./','')
PROGRAM_PID  = os.getpid()
try:
        USER_NAME = os.getlogin()
except:
        USER_NAME = pwd.getpwuid(os.getuid())[0]


__author__ = "Larne Pekowsky <lppekows@physics.syr.edu>"
__version__ = "git id %s" % git_version.id
__date__ = git_version.date


#
# =============================================================================
#
#                                 Command Line
#
# =============================================================================
#


def parse_command_line():
    """
    Parse the command line, return an options object
    """

    parser = OptionParser(
        version = "Name: %%prog\n%s" % git_version.verbose_msg,
        usage   = "%prog -v|--veto-file filename [options]",
        description = "Reads one or more segment files and a veto file and generates files of veto segments"        
	)
    
    parser.add_option("-v", "--veto-file",    metavar = "veto_file",    help = "veto XML file (required).")
    parser.add_option("-o", "--output-dir",   metavar = "output_dir",   default = '.',          help = "Directory to write output (default=cwd).")
    parser.add_option("-k", "--keep-db",      metavar = "keep_db",      action  = "store_true", help = "Keep sqlite database.")
    parser.add_option("-r", "--remove-definer-table",    action = "store_true", help = "If set the veto definer table will not be included in the output.")

    parser.add_option("-t", "--segment-url",  metavar = "segment_url", help = "Segment URL")
    parser.add_option("-d", "--database",     metavar = "use_database", action = "store_true", help = "use database specified by environment variable S6_SEGMENT_SERVER")
    parser.add_option("-f", "--dmt-file",     metavar = "use_files", action = "store_true", help = "use files in directory specified by environment variable ONLINEDQ")
    parser.add_option("-c", "--cumulative-categories",   action = "store_true", help = "If set the category N files will contain all segments in categories <= N")
    parser.add_option("-p", "--separate-categories",     action = "store_true", help = "If set the category N files will contain only category N")
    parser.add_option("-i", "--individual-results",      action = "store_true", help = "If set include information on each veto individually.")
    parser.add_option("-C", "--veto-categories", action="store", default=None, help="A comma separated list of all veto categories to consider. If not given then consider all categories in the veto-definer file.")
    parser.add_option("-I", "--ifo-list", action="store", default=None, help="A comma separated list of all ifos to consider. If not given then consider all ifos in the veto-definer file.")

    # Time options
    parser.add_option("-s", "--gps-start-time", metavar = "gps_start_time", help = "Start of GPS time range")
    parser.add_option("-e", "--gps-end-time",   metavar = "gps_end_time", help = "End of GPS time range")

    options, others = parser.parse_args()

    if not options.veto_file:
        raise ValueError("missing required argument --veto-file")
   
    # User must specify how to treat the categories
    if not (options.cumulative_categories or options.separate_categories) or (options.cumulative_categories and options.separate_categories):
        print("Must provide one of --cumulative-categories | --separate-categories", file=sys.stderr)
        sys.exit(-1)

    tmp_dir = None

    # If the veto file is a URL, load it
    loaded_file = None

    if options.veto_file.find(':') != -1:
        tmp_dir = tempfile.mkdtemp()
        inurl   = urllib.urlopen(options.veto_file)

        # Grab the part of the name after the last slash
        pos     = options.veto_file[::-1].find('/')
        fname   = (pos > -1) and options.veto_file[ -1 * pos:] or "veto_definer.xml" 

        loaded_file = tmp_dir + "/" + fname

        outfile = open(loaded_file, 'w')
        for l in inurl:
            print(l, end=' ', file=outfile)

        inurl.close()
        outfile.close()

    # Make sure we have required arguments
    database_location = None
    file_location     = None

    # Make sure we know who to contact for data
    # (This is redundant with ligolw_segment_query, maybe move to
    # query_engine)
    if options.segment_url:
        if options.segment_url.startswith('ldbd') or options.segment_url.startswith('http'):
            database_location = options.segment_url
        elif options.segment_url.startswith('file:'):
            file_location = options.segment_url[len('file://'):]
        else:
            tmp_dir = tempfile.mkdtemp()
        
            # Grab the part of the name after the last slash
            pos     = options.segment_url[::-1].find('/')
            fname   = (pos > -1) and options.segment_url[ -1 * pos:] or "dmt.xml" 

            inurl   = urllib.urlopen(options.segment_url)
            outfile = open(tmp_dir + "/" + fname, 'w')
            for l in inurl:
                print(l, end=' ', file=outfile)

            inurl.close()
            outfile.close()
            file_location = tmp_dir
    elif options.database:
        if 'S6_SEGMENT_SERVER' not in os.environ:
            raise ValueError( "--database specified by S6_SEGMENT_SERVER not set" )
        database_location = os.environ['S6_SEGMENT_SERVER']
    elif options.dmt_file:
        if 'ONLINEDQ' not in os.environ:
            raise ValueError( "--dmt-file specified but ONLINEDQ not set" )

        tmp = os.environ['ONLINEDQ']
        if tmp.startswith('file://'):
            tmp = tmp[len('file://'):]
        file_location = tmp
    else:
        raise ValueError( "One of [ --segment_url | --database | --dmt-file ] must be provided" )


    if not options.gps_start_time:
        raise ValueError( "missing required argument --gps-start-time" )
    
    if not options.gps_end_time:
        raise ValueError( "missing required argument --gps-end-time" )

        
    return options, database_location, file_location, loaded_file




#
# =============================================================================
#
#                                 Library Code
#
# =============================================================================
#

#
# Do to some sort of version skew in sqlite at LHO dbtables.get_column_info
# doesn't work if and only if a new table has been created, which we do
# in order to make the segment_backup table.  This method is a replcement
# that works in this case.
# TODO: once the environment is fixed, remove this.
#
def kludge_get_column_info(connection, table_name):
    """
    Return an in order list of (name, type) tuples describing the
    columns in the given table.
    """
    statement = None

    table_name_str = str(table_name)

    for name,sql in connection.cursor().execute("SELECT name, sql FROM sqlite_master WHERE type == 'table'"):
        if str(name) == table_name_str:
            statement = sql

    coldefs = re.match(dbtables._sql_create_table_pattern, statement).groupdict()["coldefs"]
    return [(coldef.groupdict()["name"], coldef.groupdict()["type"]) for coldef in re.finditer(dbtables._sql_coldef_pattern, coldefs) if coldef.groupdict()["name"].upper() not in ("PRIMARY", "UNIQUE", "CHECK")]

class ContentHandler(ligolw.LIGOLWContentHandler):
    connection = None


#
# =============================================================================
#
#                                     Main
#
# =============================================================================
#

if __name__ == '__main__':
    # Settings

    dbtables.get_column_info = kludge_get_column_info

    options, db_location, file_location, loaded_file = parse_command_line()    

    now = gpstime.GpsSecondsFromPyUTC(time.time())

    # 1. Load the veto file into sqlite    
    handle, temp_db    = tempfile.mkstemp(suffix='.sqlite')
    os.close(handle)
    target          = dbtables.get_connection_filename(temp_db, None, True, False)
    veto_connection = sqlite3.connect(target)
    dbtables.idmap_sync(veto_connection)


    # Create an engine to query veto information
    veto_query_engine = query_engine.SqliteQueryEngine(veto_connection)
    xml_files         = [loaded_file or options.veto_file]

    # 2. If we're working with DMT files load them into sqlite as well
    min_start_time = int(options.gps_start_time)
    max_end_time   = int(options.gps_end_time)
    
    if file_location:
        xml_files += segmentdb_utils.get_all_files_in_range(file_location, min_start_time, max_end_time)
        segment_query_engine = veto_query_engine
    else:
        segment_connection   = segmentdb_utils.setup_database(db_location)
        segment_query_engine = query_engine.LdbdQueryEngine(segment_connection)

    ContentHandler.connection = veto_connection
    dbtables.use_in(ContentHandler)
    ligolw_sqlite.insert_from_urls(xml_files, ContentHandler)

    # If we're working with DMT files make sure we have a segment table
    if file_location:
        segmentdb_utils.ensure_segment_table(veto_connection)
    
    cli_interval = ligo.segments.segmentlist([ligo.segments.segment(int(options.gps_start_time), int(options.gps_end_time))])
            
    #
    # Get the set of veto categories and ifos for which we have segments 
    #
    if options.veto_categories:
        categories=[int(x) for x in options.veto_categories.split(",")]
    else:
        categories = [row[0] for row in veto_query_engine.query("select distinct(category) from veto_definer")]
    if options.ifo_list:
        ifos = [x for x in options.ifo_list.split(",")]
    else:
        ifos = [row[0] for row in veto_query_engine.query("select distinct(ifo) from veto_definer")]

    category_selection_operator = options.cumulative_categories and ' <= ' or ' = '

    for ifo in ifos:
        for category_num in categories:
            # Start a new document
            doc = ligolw.Document()
            doc.appendChild(ligolw.LIGO_LW())

            # Copy veto definer's process table
            doc.childNodes[0].appendChild(dbtables.get_xml(veto_connection, ['process']).childNodes[0])

            # Register ourselves
            proc_id = process.register_to_xmldoc(doc, PROGRAM_NAME, options.__dict__, version=git_version.id).process_id

            # Keep the veto_definer for tracking purposes
            if not options.remove_definer_table:
                doc.childNodes[0].appendChild(dbtables.get_xml(veto_connection, ['veto_definer']).childNodes[0])

            # loop over select of name, version where ifo=current ifo and category <= current category
            cursor = veto_connection.cursor()
            sql = """SELECT ifo, name, version, start_time, end_time, start_pad, end_pad
                       FROM veto_definer
                      WHERE ifo = '%s' AND category %s %s""" % (ifo, category_selection_operator, category_num)

            segdefs = []
            for ifo, name, version, start_time, end_time, start_pad, end_pad in cursor.execute(sql):
                if end_time == 0:
                    end_time = now

                veto_interval  = ligo.segments.segmentlist([ligo.segments.segment(start_time, end_time)])
                veto_interval.coalesce()
                veto_interval &= cli_interval


                if len(veto_interval) > 0:
                    segdefs.append( (ifo, name, version, veto_interval[0][0], veto_interval[0][1], start_pad, end_pad) )


            vetoed_segments   = segmentdb_utils.query_segments(segment_query_engine, 'segment', segdefs)
            segment_summaries = segmentdb_utils.query_segments(segment_query_engine, 'segment_summary', segdefs)

            # We could write out everything we found.  This might help tracking down why a time
            # or trigger was vetoed.
            if options.individual_results:
              segmentdb_utils.add_segment_info(doc, proc_id, segdefs, vetoed_segments, segment_summaries)

            # Form the result
            vetoed_segments = reduce(operator.or_, vetoed_segments).coalesce()

            # Add the result type to the segment definer table
            seg_name   = 'VETO_CAT%d%s' % (category_num, options.cumulative_categories and '_CUMULATIVE' or '')
            seg_def_id = segmentdb_utils.add_to_segment_definer(doc, proc_id, ifo, seg_name, 1)

            # and segment summary
            segmentdb_utils.add_to_segment_summary(doc, proc_id, seg_def_id, [[min_start_time, max_end_time]])

            # and store the segments
            segmentdb_utils.add_to_segment(doc, proc_id, seg_def_id, vetoed_segments)
            
            # Dump to XML
            output_name = "%s/%s-VETOTIME_CAT%d-%d-%d.xml" % (options.output_dir, ifo, category_num, min_start_time, (max_end_time - min_start_time))
            utils.write_filename(doc, output_name)

    # Clean up
    if loaded_file:
        os.remove(loaded_file)
        pos   = loaded_file[::-1].find('/') + 1
        os.rmdir(loaded_file[:-1 * pos])


    if not options.keep_db:
        os.remove(temp_db)




#
# Copyright (C) 2006  Nickolas Fotopoulos, Chad Hanna
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
%prog [options] dax_file site

dax_file: DAX file to submit
site: site at which to run DAX; choose from {CIT, LHO, LLO, NEMO}

Specifying --ini-file is probably required unless you already have a tc.data
for this workflow.
"""

from __future__ import print_function
import six.moves.configparser
import optparse
import os
import socket
import sys
import urllib
import urlparse
import subprocess

default_sites = os.environ["GLUE_PREFIX"] + "/etc/ldg-sites.xml"
default_properties = os.environ["GLUE_PREFIX"] + "/etc/pegasus-properties.bundle"

def get_install_type(path):
  """
  Function to attempt to figure out what a binary is.  This is a
  non issue since we don't use this pegasus feauture yet
  """
  INSTALL = 'INTEL32::LINUX'
  TYPE = 'STATIC_BINARY'
  pin, pout, perr = os.popen3('file '+path)
  pid, status = os.wait()
  output = pout.readlines()[0].strip()
  if output.find('64-bit') !=-1: 
     INSTALL = 'INTEL64::LINUX'  
#  if output.find('dynamically') !=-1: 
#     TYPE = 'DYNAMIC_BINARY'
#  if output.find('script') !=-1: 
#     TYPE = 'INSTALLED'
  return INSTALL,TYPE

class PegasusTransformationCatalog(list):
    def write_tc_data(self, file_obj):
        file_obj.write("\n".join(map(str, self)) + "\n")
    
    def from_hipe_ini(cls, ini_fileobj):
        cp = six.moves.configparser.ConfigParser()
        cp.readfp(ini_fileobj)
        hostname = socket.getfqdn()

        catalog = cls()
        for prog, path in cp.items("condor"):
            if prog !="universe":
              tx = PegasusTransformation()
              tx.siteID = "local"
              tx.LogicalTX = "ligo::%s:1.0" % os.path.split(path)[1]
              tx.PhysicalTX = urlparse.urlunsplit(("gsiftp", hostname,
                  os.path.abspath(path), None, None))
              tx.SystemInfo,tx.Type = get_install_type(path) 
              #tx.Type = "STATIC_BINARY"
              #tx.SystemInfo = "INTEL64::LINUX"
              tx.Profiles = ""
              catalog.append(tx)
        return catalog
    from_hipe_ini = classmethod(from_hipe_ini)

class PegasusTransformation(object):
    __slots__ = ["siteID", "LogicalTX", "PhysicalTX", "Type", "SystemInfo",
        "Profiles"]
    
    def __str__(self):
        return " ".join([getattr(self, key) for key in self.__slots__])

def parse_args():
    parser = optparse.OptionParser(version="$Id$", usage=__doc__)
    parser.add_option("-s", "--sites-file", default=default_sites,
        help="site.xml describing the desired sites (a default is installed with glue)")
    parser.add_option("-p", "--properties-file", default=default_properties,
        help="property bundle with pegasus preferences (a default is installed with glue)")
    parser.add_option("-i", "--ini-file",
        help="inspiral_hipe's ini file with which to create tc.data")
    parser.add_option("-n", "--no-submit", action="store_true", default=False,
        help="concretize the DAG and generate a sub file for it; do not submit to cluster")
    parser.add_option("-c", "--pegasus-cache",
        help="Point to a pegasus RLS cache file (if inputs aren't in catalog)")
    parser.add_option("-v", "--verbose", action="store_true", default=False,
        help="print informative messages")
    
    opts, args = parser.parse_args()
    
    if len(args) < 2:
        print("not enough arguments", file=sys.stderr)
    
    return opts, args

#
# Main
#

# commandline
opts, args = parse_args()
dax_file = args[0]
site = args[1]

# create tc.data if necessary
if opts.ini_file:
    catalog = PegasusTransformationCatalog.from_hipe_ini(open(opts.ini_file))
    catalog.write_tc_data(open("tc.data", "w"))

#
# run Pegasus stuff
#

# do pegasus-plan
cwd = os.getcwd()

plan_args = ["pegasus-plan"]
plan_args.append("-Dpegasus.catalog.site.file=" + opts.sites_file)
plan_args.append("-Dpegasus.user.properties=" + opts.properties_file)
plan_args.append("-Dpegasus.catalog.transformation.file=" + os.path.abspath("tc.data"))
plan_args.append("-Dpegasus.dir.storage=%s" % cwd)
plan_args.append("--relative-dir=./")
plan_args.append("--sites=" + site)
plan_args.append("--output local")
plan_args.append("--dax=" + dax_file)
#plan_args.append("--nocleanup")
plan_args.append("--basename="+dax_file)
plan_args.append("--job-prefix="+dax_file)
#plan_args.append("--verbose")
plan_args.append("--randomdir")
plan_args.append("--force")
if not opts.no_submit:
  plan_args.append("--submit")

if opts.pegasus_cache: plan_args.append('--cache='+opts.pegasus_cache)
plan_command = " ".join(plan_args)

if opts.verbose: print("Executing: " + plan_command)
sto = open(".ldg_submit_dax.stdout","w");
ste = open(".ldg_submit_dax.stderr","w");


popen = subprocess.Popen(plan_command.split(), stderr = ste, stdout = sto)
(stdout, stderr) = popen.communicate()
status = popen.returncode
sto.close()
ste.close()

if status != 0:
    print("External call failed.", file=sys.stderr)
    print("  status: %d" % status, file=sys.stderr)
    print("  stdout: %s" % open(".ldg_submit_dax.stdout","r").readlines(), file=sys.stderr)
    print("  stderr: %s" % open(".ldg_submit_dax.stderr","r").readlines(), file=sys.stderr)
    print("  command: %s" % plan_command, file=sys.stderr)
    sys.exit(1) 



#
#
# Copyright (C) 2009  Larne Pekowsky
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


#
# =============================================================================
#
#                                   Preamble
#
# =============================================================================
#


"""
Displays the set of flags active at a give time or times, in a convinent
ascii form.

All the heavy lifting is done by ligolw_dq_query --segment-url, this just 
processes the arguments and output.
"""

from __future__ import print_function
from optparse import OptionParser

import sys
import os
import glob
import time
import socket
import tempfile
import urllib
import pwd

from glue import git_version

PROGRAM_NAME = sys.argv[0].replace('./','')
PROGRAM_PID  = os.getpid()
USER_NAME    = pwd.getpwuid(os.getuid())[0]



__author__  = "Larne Pekowsky <lppekows@physics.syr.edu>"
__version__ = "git id %s" % git_version.id
__date__ = git_version.date

#
# =============================================================================
#
#                                 Command Line
#
# =============================================================================
#


def parse_command_line():
    """
    Parse the command line, return an options object
    """

    parser = OptionParser(
        version = "Name: %%prog\n%s" % git_version.verbose_msg,
        usage       = "%prog --ifo ifo --time-file timefile --input-dir dir",
        description = "Prints a list of the flags that were active at each of the specified times"
	)

    # Major modes
    parser.add_option("-i", "--ifo",        metavar = "ifo",       help = "Abbreviation for IFO of interest")
    parser.add_option("-t", "--time-file",  metavar = "time-file", help = "File of gps times, one per line")
    parser.add_option("-s", "--segment-url", metavar = "segment-url", help = "URL of segment database or DMT files")
    parser.add_option("-d", "--input-dir",  metavar = "input-dir", help = "Location of DMT files")


    options, others = parser.parse_args()


    if not options.ifo:
        print("ifo is required", file=sys.stderr)
        sys.exit(1)



    if not options.time_file:
        print("time-file is required", file=sys.stderr)
        sys.exit(1)

    if not os.path.exists(options.time_file):
        print("time-file %s does not exist" % options.time_file, file=sys.stderr)
        sys.exit(1)

    if not os.path.isfile(options.time_file):
        print("time-file %s exists but is not a file" % options.time_file, file=sys.stderr)
        sys.exit(1)




    if not (options.input_dir or options.segment):
        print("One of [input-dir|segment] is required", file=sys.stderr)
        sys.exit(1)

    if options.input_dir and not os.path.exists(options.input_dir):
        print("input_dir %s does not exist" % options.input_dir, file=sys.stderr)
        sys.exit(1)

    if options.input_dir and not os.path.isdir(options.input_dir):
        print("input_dir %s exists but is not a directory" % options.input_dir, file=sys.stderr)
        sys.exit(1)


    return options




#
# =============================================================================
#
#                                     Main
#
# =============================================================================
#

if __name__ == '__main__':
    options = parse_command_line()    

    f = open(options.time_file)

    if options.input_dir:
        segment = 'file://' + options.input_dir
    else:
        segment = options.segment

    for tme in f:
        tme  = tme[:-1]
        pipe = os.popen('ligolw_dq_query --segment=%s --active --include-segments=%s %s | ligolw_print -t segment_definer -c name' % (segment, options.ifo, tme))

        print(tme, end=' ')

        for name in pipe:
            print('\t', name[:-1], end=' ')

        pipe.close()

        print()

    f.close()




# Copyright (C) 2006--2014  Kipp Cannon
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


#
# =============================================================================
#
#                                   Preamble
#
# =============================================================================
#


"""
Library of utility code for LIGO Light Weight XML applications.
"""


import codecs
import gzip
from hashlib import md5
import warnings
import os
from six.moves import urllib
import signal
import stat
import sys


from glue import git_version
from .. import ligolw


__author__ = "Kipp Cannon <kipp.cannon@ligo.org>"
__version__ = "git id %s" % git_version.id
__date__ = git_version.date


__all__ = ["sort_files_by_size", "local_path_from_url", "load_fileobj", "load_filename", "load_url", "write_fileobj", "write_filename", "write_url"]


#
# =============================================================================
#
#                                 Input/Output
#
# =============================================================================
#


def sort_files_by_size(filenames, verbose = False, reverse = False):
	"""
	Return a list of the filenames sorted in order from smallest file
	to largest file (or largest to smallest if reverse is set to True).
	If a filename in the list is None (used by many glue.ligolw based
	codes to indicate stdin), its size is treated as 0.  The filenames
	may be any sequence, including generator expressions.
	"""
	if verbose:
		if reverse:
			sys.stderr.write("sorting files from largest to smallest ...\n")
		else:
			sys.stderr.write("sorting files from smallest to largest ...\n")
	return sorted(filenames, key = (lambda filename: os.stat(filename)[stat.ST_SIZE] if filename is not None else 0), reverse = reverse)


def local_path_from_url(url):
	"""
	For URLs that point to locations in the local filesystem, extract
	and return the filesystem path of the object to which they point.
	As a special case pass-through, if the URL is None, the return
	value is None.  Raises ValueError if the URL is not None and does
	not point to a local file.

	Example:

	>>> print(local_path_from_url(None))
	None
	>>> local_path_from_url("file:///home/me/somefile.xml.gz")
	'/home/me/somefile.xml.gz'
	"""
	if url is None:
		return None
	scheme, host, path = urllib.parse.urlparse(url)[:3]
	if scheme.lower() not in ("", "file") or host.lower() not in ("", "localhost"):
		raise ValueError("%s is not a local file" % repr(url))
	return path


class RewindableInputFile(object):
	# The GzipFile class in Python's standard library is, in my
	# opinion, somewhat weak.  Instead of relying on the return values
	# from the file object's .read() method, GzipFile checks for EOF
	# using calls to .seek().  Furthermore, it uses .seek() instead of
	# buffering data internally as required.  This makes GzipFile
	# gratuitously unable to work with pipes, urlfile objects, and
	# anything else that does not support seeking (including the
	# MD5File class in this module).  To hack around this, this class
	# provides the buffering needed by GzipFile.  It also does proper
	# EOF checking, and uses the results to emulate the results of
	# GzipFile's .seek() games.
	#
	# By wrapping your file object in this class before passing it to
	# GzipFile, you can use GzipFile to read from non-seekable files.
	#
	# How GzipFile checks for EOF == call .tell() to get current
	# position, seek to end of file with .seek(0, 2), call .tell()
	# again and check if the number has changed from before, if it has
	# then we weren't at EOF so call .seek() with original position and
	# keep going.  ?!

	def __init__(self, fileobj, buffer_size = 1024):
		# the real source of data
		self.fileobj = fileobj
		# where the application thinks it is in the file.  this is
		# used to fake .tell() because file objects that don't
		# support seeking, like stdin, report IOError, and the
		# things returned by urllib don't have a .tell() method at
		# all.
		self.pos = 0
		# how many octets of the internal buffer to return before
		# getting more data
		self.reuse = 0
		# the internal buffer
		self.buf = b' ' * buffer_size
		# flag indicating a .seek()-based EOF test is in progress
		self.gzip_hack_pretend_to_be_at_eof = False
		# avoid attribute look-ups
		self._next = self.fileobj.next
		self._read = self.fileobj.read
		self.close = self.fileobj.close

	def __iter__(self):
		return self

	def next(self):
		if self.gzip_hack_pretend_to_be_at_eof:
			return b''
		if self.reuse:
			buf = self.buf[-self.reuse:]
			self.reuse = 0
		else:
			buf = self._next()
			self.buf = (self.buf + buf)[-len(self.buf):]
		self.pos += len(buf)
		return buf

	def read(self, size = None):
		if self.gzip_hack_pretend_to_be_at_eof:
			return b''
		if self.reuse:
			if self.reuse < 0:
				buf = self._read(size - self.reuse)
				self.buf = (self.buf + buf)[-len(self.buf):]
				buf = buf[-self.reuse:]
				self.reuse = 0
			# size is None --> condition is False
			elif 0 <= size < self.reuse:
				buf = self.buf[-self.reuse:-self.reuse + size]
				self.reuse -= size
			else:
				buf = self.buf[-self.reuse:]
				self.reuse = 0
				# size is None --> condition is False
				if len(buf) < size:
					buf += self.read(size - len(buf))
		else:
			buf = self._read(size)
			self.buf = (self.buf + buf)[-len(self.buf):]
		self.pos += len(buf)
		return buf

	def seek(self, offset, whence = os.SEEK_SET):
		self.gzip_hack_pretend_to_be_at_eof = False
		if whence == os.SEEK_SET:
			if offset >= 0 and 0 <= self.pos + self.reuse - offset < len(self.buf):
				self.reuse += self.pos - offset
				self.pos = offset
			else:
				raise IOError("seek out of range")
		elif whence == os.SEEK_CUR:
			if self.reuse - len(self.buf) <= offset:
				self.reuse -= offset
				self.pos += offset
			else:
				raise IOError("seek out of range")
		elif whence == os.SEEK_END:
			if offset == 0:
				self.gzip_hack_pretend_to_be_at_eof = True
			else:
				raise IOError("seek out of range")

	def tell(self):
		if self.gzip_hack_pretend_to_be_at_eof:
			# check to see if we are at EOF by seeing if we can
			# read 1 character.  save it in the internal buffer
			# to not loose it.
			c = self._read(1)
			self.buf = (self.buf + c)[-len(self.buf):]
			self.reuse += len(c)
			if c:
				# this will not return the same answer as
				# when GzipFile called it before seeking to
				# EOF
				return self.pos + 1
		return self.pos

	def __enter__(self):
		return self

	def __exit__(self, *args):
		self.close()
		return False


class MD5File(object):
	def __init__(self, fileobj, md5obj = None, closable = True):
		self.fileobj = fileobj
		if md5obj is None:
			self.md5obj = md5()
		else:
			self.md5obj = md5obj
		self.closable = closable
		# avoid attribute look-ups
		self._update = self.md5obj.update
		try:
			self._next = self.fileobj.next
		except AttributeError:
			# replace our .next() method with something that
			# will raise a more meaningful exception if
			# attempted
			self.next = lambda *args, **kwargs: fileobj.next(*args, **kwargs)
		try:
			self._read = self.fileobj.read
		except AttributeError:
			# replace our .read() method with something that
			# will raise a more meaningful exception if
			# attempted
			self.read = lambda *args, **kwargs: fileobj.read(*args, **kwargs)
		try:
			self._write = self.fileobj.write
		except AttributeError:
			# replace our .write() method with something that
			# will raise a more meaningful exception if
			# attempted
			self.write = lambda *args, **kwargs: fileobj.write(*args, **kwargs)
		try:
			self.tell = self.fileobj.tell
		except AttributeError:
			self.tell = lambda *args, **kwargs: fileobj.tell(*args, **kwargs)
		try:
			self.flush = self.fileobj.flush
		except AttributeError:
			self.flush = lambda *args, **kwargs: fileobj.flush(*args, **kwargs)

	def __iter__(self):
		return self

	def next(self):
		buf = self._next()
		self._update(buf)
		return buf

	def read(self, *args):
		buf = self._read(*args)
		self._update(buf)
		return buf

	def write(self, buf):
		self._update(buf)
		return self._write(buf)

	def close(self):
		if self.closable:
			return self.fileobj.close()
		else:
			# at least make sure we're flushed
			self.flush()

	def __enter__(self):
		return self

	def __exit__(self, *args):
		self.close()
		return False


class SignalsTrap(object):
	default_signals = (signal.SIGTERM, signal.SIGTSTP)

	def __init__(self, trap_signals = default_signals):
		self.trap_signals = trap_signals

	def handler(self, signum, frame):
		self.deferred_signals.append(signum)

	def __enter__(self):
		self.oldhandlers = {}
		self.deferred_signals = []
		if self.trap_signals is None:
			return self
		for sig in self.trap_signals:
			self.oldhandlers[sig] = signal.getsignal(sig)
			signal.signal(sig, self.handler)
		return self

	def __exit__(self, *args):
		# restore original handlers
		for sig, oldhandler in self.oldhandlers.items():
			signal.signal(sig, oldhandler)
		# send ourselves the trapped signals in order
		while self.deferred_signals:
			os.kill(os.getpid(), self.deferred_signals.pop(0))
		return False


def load_fileobj(fileobj, gz = None, xmldoc = None, contenthandler = None):
	"""
	Parse the contents of the file object fileobj, and return the
	contents as a LIGO Light Weight document tree.  The file object
	does not need to be seekable.

	If the gz parameter is None (the default) then gzip compressed data
	will be automatically detected and decompressed, otherwise
	decompression can be forced on or off by setting gz to True or
	False respectively.

	If the optional xmldoc argument is provided and not None, the
	parsed XML tree will be appended to that document, otherwise a new
	document will be created.  The return value is a tuple, the first
	element of the tuple is the XML document and the second is a string
	containing the MD5 digest in hex digits of the bytestream that was
	parsed.

	Example:

	>>> from glue.ligolw import ligolw
	>>> from io import BytesIO
	>>> f = BytesIO(b'<?xml version="1.0" encoding="utf-8" ?><!DOCTYPE LIGO_LW SYSTEM "http://ldas-sw.ligo.caltech.edu/doc/ligolwAPI/html/ligolw_dtd.txt"><LIGO_LW><Table Name="demo:table"><Column Name="name" Type="lstring"/><Column Name="value" Type="real8"/><Stream Name="demo:table" Type="Local" Delimiter=",">"mass",0.5,"velocity",34</Stream></Table></LIGO_LW>')
	>>> xmldoc, digest = load_fileobj(f, contenthandler = ligolw.LIGOLWContentHandler)
	>>> digest
	'6bdcc4726b892aad913531684024ed8e'

	The contenthandler argument specifies the SAX content handler to
	use when parsing the document.  The contenthandler is a required
	argument.  See the glue.ligolw package documentation for typical
	parsing scenario involving a custom content handler.  See
	glue.ligolw.ligolw.PartialLIGOLWContentHandler and
	glue.ligolw.ligolw.FilteringLIGOLWContentHandler for examples of
	custom content handlers used to load subsets of documents into
	memory.
	"""
	fileobj = MD5File(fileobj)
	md5obj = fileobj.md5obj
	if gz or gz is None:
		fileobj = RewindableInputFile(fileobj)
		magic = fileobj.read(2)
		fileobj.seek(0, os.SEEK_SET)
		if gz or magic == b'\037\213':
			fileobj = gzip.GzipFile(mode = "rb", fileobj = fileobj)
	if xmldoc is None:
		xmldoc = ligolw.Document()
	ligolw.make_parser(contenthandler(xmldoc)).parse(fileobj)
	return xmldoc, md5obj.hexdigest()


def load_filename(filename, verbose = False, **kwargs):
	"""
	Parse the contents of the file identified by filename, and return
	the contents as a LIGO Light Weight document tree.  stdin is parsed
	if filename is None.  Helpful verbosity messages are printed to
	stderr if verbose is True.  All other keyword arguments are passed
	to load_fileobj(), see that function for more information.  In
	particular note that a content handler must be specified.

	Example:

	>>> from glue.ligolw import ligolw
	>>> xmldoc = load_filename("demo.xml", contenthandler = ligolw.LIGOLWContentHandler, verbose = True)
	"""
	if verbose:
		sys.stderr.write("reading %s ...\n" % (("'%s'" % filename) if filename is not None else "stdin"))
	if filename is None:
		# In Python 3, ``sys.stdin`` has an attribute called
		# ``buffer`` that is the underyling byte-oriented stream.
		xmldoc, hexdigest = load_fileobj(sys.stdin.buffer if hasattr(sys.stdin, "buffer") else sys.stdin, **kwargs)
	else:
		with open(filename, "rb") as fileobj:
			xmldoc, hexdigest = load_fileobj(fileobj, **kwargs)
	if verbose:
		sys.stderr.write("md5sum: %s  %s\n" % (hexdigest, (filename if filename is not None else "")))
	return xmldoc


def load_url(url, verbose = False, **kwargs):
	"""
	Parse the contents of file at the given URL and return the contents
	as a LIGO Light Weight document tree.  Any source from which
	Python's urllib library can read data is acceptable.  stdin is
	parsed if url is None.  Helpful verbosity messages are printed to
	stderr if verbose is True.  All other keyword arguments are passed
	to load_fileobj(), see that function for more information.  In
	particular note that a content handler must be specified.

	Example:

	>>> from os import getcwd
	>>> from glue.ligolw import ligolw
	>>> xmldoc = load_url("file://localhost/%s/demo.xml" % getcwd(), contenthandler = ligolw.LIGOLWContentHandler, verbose = True)
	"""
	if verbose:
		sys.stderr.write("reading %s ...\n" % (("'%s'" % url) if url is not None else "stdin"))
	if url is None:
		# In Python 3, ``sys.stdin`` has an attribute called
		# ``buffer`` that is the underyling byte-oriented stream.
		xmldoc, hexdigest = load_fileobj(sys.stdin.buffer if hasattr(sys.stdin, "buffer") else sys.stdin, **kwargs)
	else:
		scheme, host, path = urllib.parse.urlparse(url)[:3]
		if scheme.lower() in ("", "file") and host.lower() in ("", "localhost"):
			fileobj = lambda: open(path, "rb")
		else:
			fileobj = lambda: urllib.request.urlopen(url)
		with fileobj() as fileobj:
			xmldoc, hexdigest = load_fileobj(fileobj, **kwargs)
	if verbose:
		sys.stderr.write("md5sum: %s  %s\n" % (hexdigest, (url if url is not None else "")))
	return xmldoc


def write_fileobj(xmldoc, fileobj, gz = False, compresslevel = 3, **kwargs):
	"""
	Writes the LIGO Light Weight document tree rooted at xmldoc to the
	given file object.  Internally, the .write() method of the xmldoc
	object is invoked and any additional keyword arguments are passed
	to that method.  The file object need not be seekable.  The output
	data is gzip compressed on the fly if gz is True, and in that case
	the compresslevel parameter sets the gzip compression level (the
	default is 3).  The return value is a string containing the hex
	digits of the MD5 digest of the output bytestream.

	Example:

	>>> from io import BytesIO
	>>> import sys
	>>> from glue.ligolw import ligolw
	>>> xmldoc = load_filename("demo.xml", contenthandler = ligolw.LIGOLWContentHandler)
	>>> f = BytesIO()
	>>> digest = write_fileobj(xmldoc, f)
	>>> print(f.getvalue().decode())	# doctest: +NORMALIZE_WHITESPACE
	<?xml version='1.0' encoding='utf-8'?>
	<!DOCTYPE LIGO_LW SYSTEM "http://ldas-sw.ligo.caltech.edu/doc/ligolwAPI/html/ligolw_dtd.txt">
	<LIGO_LW>
		<Table Name="demo:table">
			<Column Name="name" Type="lstring"/>
			<Column Name="value" Type="real8"/>
			<Stream Delimiter="," Name="demo:table" Type="Local">
	"mass",0.5,"velocity",34
			</Stream>
		</Table>
	</LIGO_LW>
	>>> digest
	'41503b904a55576e6430aeba8b14dac6'
	"""
	with MD5File(fileobj, closable = False) as fileobj:
		md5obj = fileobj.md5obj
		with fileobj if not gz else gzip.GzipFile(mode = "wb", fileobj = fileobj, compresslevel = compresslevel) as fileobj:
			with codecs.getwriter("utf_8")(fileobj) as fileobj:
				xmldoc.write(fileobj, **kwargs)
		return md5obj.hexdigest()


class tildefile(object):
	def __init__(self, filename):
		if not filename:
			raise ValueError(filename)
		self.filename = filename

	def __enter__(self):
		try:
			self.tildefilename = self.filename + "~"
			self.fobj = open(self.tildefilename, "wb")
		except IOError:
			self.tildefilename = None
			self.fobj = open(self.filename, "wb")
		return self.fobj

	def __exit__(self, exc_type, exc_val, exc_tb):
		self.fobj.close()
		del self.fobj

		#
		# only rename the "~" version to the final destination if
		# no exception has occurred.
		#

		if exc_type is None and self.tildefilename is not None:
			os.rename(self.tildefilename, self.filename)

		return False


def write_filename(xmldoc, filename, verbose = False, gz = False, with_mv = True, trap_signals = SignalsTrap.default_signals, **kwargs):
	"""
	Writes the LIGO Light Weight document tree rooted at xmldoc to the
	file name filename.  If filename is None the file is written to
	stdout, otherwise it is written to the named file.  Friendly
	verbosity messages are printed while writing the file if verbose is
	True.  The output data is gzip compressed on the fly if gz is True.
	If with_mv is True and filename is not None the filename has a "~"
	appended to it and the file is written to that name then moved to
	the requested name once the write has completed successfully.

	Internally, write_fileobj() is used to perform the write.  All
	additional keyword arguments are passed to write_fileobj().

	This function traps the signals in the trap_signals iterable during
	the write process (see SignalsTrap for the default signals), and it
	does this by temporarily installing its own signal handlers in
	place of the current handlers.  This is done to prevent Condor
	eviction during the write process.  When the file write is
	concluded the original signal handlers are restored.  Then, if
	signals were trapped during the write process, the signals are then
	resent to the current process in the order in which they were
	received.  The signal.signal() system call cannot be invoked from
	threads, and trap_signals must be set to None or an empty sequence
	if this function is used from a thread.

	Example:

	>>> write_filename(xmldoc, "demo.xml")	# doctest: +SKIP
	>>> write_filename(xmldoc, "demo.xml.gz", gz = True)	# doctest: +SKIP
	"""
	if verbose:
		sys.stderr.write("writing %s ...\n" % (("'%s'" % filename) if filename is not None else "stdout"))
	with SignalsTrap(trap_signals):
		if filename is None:
			# In Python 3, ``sys.stdout`` has an attribute
			# called ``buffer`` that is the underyling
			# byte-oriented stream.
			fileobj = sys.stdout
			try:
				fileobj = fileobj.buffer
			except AttributeError:
				pass
			hexdigest = write_fileobj(xmldoc, fileobj, gz = gz, **kwargs)
		else:
			if not gz and filename.endswith(".gz"):
				warnings.warn("filename '%s' ends in '.gz' but file is not being gzip-compressed" % filename, UserWarning)
			binary_open = lambda filename: open(filename, 'wb')
			with (binary_open if not with_mv else tildefile)(filename) as fileobj:
				hexdigest = write_fileobj(xmldoc, fileobj, gz = gz, **kwargs)
	if verbose:
		sys.stderr.write("md5sum: %s  %s\n" % (hexdigest, (filename if filename is not None else "")))


def write_url(xmldoc, url, **kwargs):
	"""
	Writes the LIGO Light Weight document tree rooted at xmldoc to the
	URL name url.

	NOTE:  only URLs that point to local files can be written to at
	this time.  Internally, write_filename() is used to perform the
	write.  All additional keyword arguments are passed to that
	function.  The implementation might change in the future,
	especially if support for other types of URLs is ever added.

	Example:

	>>> write_url(xmldoc, "file:///data.xml")	# doctest: +SKIP
	>>> write_url(xmldoc, "file:///data.xml.gz", gz = True)	# doctest: +SKIP
	"""
	return write_filename(xmldoc, local_path_from_url(url), **kwargs)



#
#
# Copyright (C) 2009  Larne Pekowsky, Ping Wei
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


#
# =============================================================================
#
#                                   Preamble
#
# =============================================================================
#


"""
Provides tools to query the segment database or DMT XML files regarding the
set of flags defined/active at or near a given time:


The --report option prints an extensive query of the database for (a) given gps
time(s). It prints which flags were undefined at the time (using the
information in the segment_summary table). For the flags which were defined, it
determines if the flag was active or inactive at that time. For an active flag,
it prints the start and end time of the segment to which the active. For an
inactive flag, it prints the end time of the previous adjacent active segment
and the start time of the next adjacent active segment.

  * What is the status of all DQ flags at this time? ligolw_dq_query --report
  * Refine this query with --include-segments
  * What flags were defined at this time (i.e. query segment_summary table)? ligolw_dq_query --defined
  * Was a given flag defined at this time (i.e. query segment_summary table)? ligolw_dq_query --defined --include-segments
  * What flags were active at this time? ligolw_dq_querty --active
  * Was a given flag active at this time? ligolw_dq_query --active --include-segments
  * Near this time? ligolw_dq_query --start-pad --end-pad
"""


from __future__ import print_function
from optparse import OptionParser
import sqlite3
import sys
import os
import glob
import time
import socket
import tempfile
import urllib
import pwd

from glue.ligolw import ligolw
from glue.ligolw import table
from glue.ligolw import lsctables
from glue.ligolw import utils

from glue.ligolw.utils import ligolw_add
from glue.ligolw.utils import process
from glue.segmentdb import query_engine
from glue.segmentdb import segmentdb_utils

from glue.ligolw.utils import ligolw_sqlite
from glue.ligolw import dbtables
from glue import git_version

PROGRAM_NAME = sys.argv[0].replace('./','')
PROGRAM_PID  = os.getpid()

try:
  USER_NAME = os.getlogin()
except:
  USER_NAME = pwd.getpwuid(os.getuid())[0]




__author__  = "Larne Pekowsky <lppekows@physics.syr.edu>"
__version__ = "git id %s" % git_version.id
__date__ = git_version.date

#
# =============================================================================
#
#                                 Command Line
#
# =============================================================================
#


def parse_command_line():
    """
    Parse the command line, return an options object
    """

    parser = OptionParser(
        version = "Name: %%prog\n%s" % git_version.verbose_msg,
        usage       = "%prog [ --version | --ping | --defined | --active | --report ]  [ --segment | --database | --dmt-file ] options gps-time1 gps-time2 ... ",
        description = "Performs a number of queries against either a set of DMT files or a segment database"
	)


    # Major modes
    parser.add_option("-p", "--ping",     action = "store_true", help = "Ping the target server")
    parser.add_option("-y", "--defined",  action = "store_true", help = "Returns a segment summary table containing segments defined at the given time(s).")
    parser.add_option("-u", "--active",   action = "store_true", help = "Returns a segment table containing segments active at the given time(s).")
    parser.add_option("-q", "--report",   action = "store_true", help = "Prints which flags are defined/undefined at the given time(s). For the flags which were defined, it determines if the flag was active or inactive at that time. For an active flag, it prints the start and end time of the segment to which the active. For an inactive flag, it prints the end time of the previous adjacent active segment and the start time of the next adjacent active segment")

    # Time options
    parser.add_option("-s", "--start-pad", metavar = "start_pad", help = "Seconds before given time(s) to include in query")
    parser.add_option("-e", "--end-pad",   metavar = "end_pad",   help = "Seconds after given time(s) to include in query")


    # Data location options
    parser.add_option("-t", "--segment-url",    metavar = "segment_url", help = "Segment URL")
    parser.add_option("-d", "--database",   metavar = "use_database", action = "store_true", help = "use database specified by environment variable S6_SEGMENT_SERVER")
    parser.add_option("-f", "--dmt-files",   metavar = "use_files", action = "store_true", help = "use files in directory specified by environment variable ONLINEDQ")


    # Other options
    parser.add_option("-a", "--include-segments", metavar = "include_segments", help = "This option expects a comma separated list of a colon separated sublist of interferometer, segment type, and version. The union of segments from all types and versions specified is returned. Use --show-types to see what types are available.   For example: --include-segment-types H1:SCIENCE:1,H1:INJECTION:2 will return the segments for which H1 is in either SCIENCE version 1 or INJECTION version 2 mode. If version information is not provided, the union of the segments of the latest version of requested segment type(s) will be returned.")

    parser.add_option("-o", "--output-file",   metavar = "output_file", help = "File to which output should be written.  Defaults to stdout.")
   
    parser.add_option("-i", "--in-segments-only", action = "store_true", help = "If set, report will only return segments that given times were within") 

    options, times = parser.parse_args()

    # Make sure we have exactly one thing to do
    count = 0
    for arg in [options.ping, options.defined, options.active, options.report]:
        if arg:
            count += 1
            
    if count != 1:
        raise ValueError("Exactly one of [ --ping | --defined | --active | --report ] must be provided")
    
    
    # Make sure we have required arguments
    database_location = None
    file_location     = None

    # Make sure we know who to contact for data
    if options.segment_url:
        if options.segment_url.startswith('ldbd') or options.segment_url.startswith('http'):
            database_location = options.segment_url
        elif options.segment_url.startswith('file:'):
            file_location = options.segment_url[len('file://'):]
        else:
            tmp_dir = tempfile.mkdtemp()
        
            # Grab the part of the name after the last slash
            pos     = options.segment_url[::-1].find('/')
            fname   = (pos > -1) and options.segment_url[ -1 * pos:] or "dmt.xml" 

            inurl   = urllib.urlopen(options.segment_url)
            outfile = open(tmp_dir + "/" + fname, 'w')
            for l in inurl:
                print(l, end=' ', file=outfile)

            inurl.close()
            outfile.close()
            file_location = tmp_dir
    elif options.database:
        if 'S6_SEGMENT_SERVER' not in os.environ:
            raise ValueError( "--database specified but S6_SEGMENT_SERVER not set" )
        database_location = os.environ['S6_SEGMENT_SERVER']
    elif options.dmt_files:
        if 'ONLINEDQ' not in os.environ:
            raise ValueError( "--dmt-files specified but ONLINEDQ not set" )

        tmp = os.environ['ONLINEDQ']
        if tmp.startswith('file://'):
            tmp = tmp[len('file://'):]
        file_location = tmp
    else:
        raise ValueError( "One of [ --segment | --database | --dmt-file ] must be provided" )
        

    # Unless we're pinging, make sure we have start and end times
    if options.ping:
        if not database_location:
            raise ValueError("--ping requires [ --segment https://... | --database ]")
    
    # Report is only available when hitting the database, since 
    # we have no way of knowing where the segment boundaries are
    # and would have to load *every* XML file.
    if file_location and options.report:
        print("--report can only be used with the database.")
        sys.exit(-1)

    return options, database_location, file_location, times




#
# =============================================================================
#
#                                 General utilities
#
# =============================================================================
#

def build_time_clause(table, times):
    """Given the name of a table containing start_time and end_time columns, and an
    array of times of the form (lower bound, upper bound), constructs a SQL clause
    that restricts a search to the given ranges"""

    subclauses = ['NOT (%d > %s.end_time OR %s.start_time > %d)' % (x[0], table, table, x[1]) for x in times]
    return '( ' + " OR ".join(subclauses) + ' )'


def build_segment_clause(clauses):
    """Given an array of segment definer restrictions of the form ifo:name:version,
    ifo:name:* or ifo:name, constructs a SQL clause that restrcts a search to that
    set of definers"""

    def build_segment_subclause(clause):
        tmp = clause.split(':')

        if len(tmp) == 1 or tmp[1] == '*':
            return "segment_definer.ifos = '%s'" % (tmp[0])

        if len(tmp) == 2 or tmp[2] == '*':
            return "segment_definer.ifos = '%s' AND segment_definer.name = '%s'" % (tmp[0], tmp[1])

        return "segment_definer.ifos = '%s' AND segment_definer.name = '%s' AND segment_definer.version = %s" % (tmp[0], tmp[1], tmp[2])
        
    subclauses = list(map(build_segment_subclause, clauses.split(',')))
    return " OR ".join(subclauses)


def get_full_name(ifo, name, version):
    return '%s:%s:%d' % (ifo.strip(), name.strip(), version)

#
# =============================================================================
#
#                          Methods that implement major modes
#
# =============================================================================
#

def run_report(doc, process_id, engine, include_segments, times, in_segments_only):
    all_keys       = {}
    max_start      = {}
    max_end        = {}
    min_start      = {}
    min_end        = {}
    segment_clause = ''

    if include_segments:
        segment_clause = ' AND ' + build_segment_clause(include_segments)

    # I'm not happy with the fact that this is done with four separate
    # queries.  However it is possible that, say, there won't be a segment
    # after the given time in which case 
    #
    #   select MAX(end_time), MIN(start_time)
    #   where end_time < given time and start_time > given time
    #
    # would miss that segment definer entirely.  Maybe there's some
    # clever trick using outer joins to get around this, but I
    # can't seem to come up with one at the moment.
    #
    # For XML-file queries this isn't a problem, since local queries
    # are fast.  But talking over the LDBD server takes a noticable amount
    # of time per-query
    #
    for tm in [int(x[0]) for x in times]:
        # Segments we are in the midsts of
        rows = engine.query("""SELECT segment_definer.ifos, segment_definer.name, segment_definer.version,
             segment.start_time, segment.end_time
             FROM segment_definer, segment
             WHERE segment_definer.segment_def_id = segment.segment_def_id
             AND %d BETWEEN segment.start_time AND segment.end_time
             %s """ % (tm, segment_clause)) 

        in_times = {}
        for ifo, name, version, start_time, end_time in rows:
            full_name = get_full_name(ifo, name, version)
            in_times[full_name] = [start_time, end_time]
            print('%-45s [%d %d %d)' % (full_name, start_time, tm, end_time))

        if not in_segments_only:
            # Segments we are between. Latest end time before of interest.
            out_times = {}

            rows = engine.query("""SELECT segment_definer.ifos, segment_definer.name, segment_definer.version, MAX(segment.end_time)
                FROM segment_definer, segment
                WHERE segment_definer.segment_def_id = segment.segment_def_id
                AND segment.end_time < %d %s
                GROUP BY segment_definer.ifos, segment_definer.name, segment_definer.version""" % (tm, segment_clause))
           
           
            for ifo, name, version, end_time in rows:
                full_name = get_full_name(ifo, name, version)
                out_times[full_name] = ['%d)' % end_time, '[now']

            # The next start time after the time of interest
            rows = engine.query("""SELECT segment_definer.ifos, segment_definer.name, segment_definer.version, MIN(segment.start_time)
                FROM segment_definer, segment
                WHERE segment_definer.segment_def_id = segment.segment_def_id
                AND segment.start_time > %d %s
                GROUP BY segment_definer.ifos, segment_definer.name, segment_definer.version""" % (tm, segment_clause))

            for ifo, name, version, start_time in rows:
                full_name = get_full_name(ifo, name, version)
                value = full_name in out_times and [out_times[full_name][0], '[%s' % start_time] or ['never)', '[%s' % start_time]
                out_times[full_name] = value

            for key in out_times:
                if key not in in_times:
                    value = out_times[key]
                    print('%-45s %s %d %s' % (key, value[0], tm, value[1]))


def run_active(doc, process_id, engine, include_segments, times):
    time_clause    = build_time_clause('segment', times)
    segment_clause = ''

    if include_segments:
        segment_clause = ' AND ' + build_segment_clause(include_segments)

    rows = engine.query("""SELECT segment_definer.ifos, segment_definer.name, segment_definer.version, segment_definer.comment, segment.start_time, segment.end_time
        FROM segment_definer, segment_summary, segment
        WHERE segment_definer.segment_def_id = segment_summary.segment_def_id
        AND   segment.start_time BETWEEN segment_summary.start_time AND segment_summary.end_time
        AND   segment_definer.segment_def_id = segment.segment_def_id
        AND """ + time_clause + segment_clause)

    
    distinct_names = []
    distinct_rows = []
    for x in rows:
      if x[0:3] not in distinct_names:
         distinct_names.append(x[0:2])
         distinct_rows.append(x)


    seg_def_table = lsctables.New(lsctables.SegmentDefTable, columns = ["process_id", "segment_def_id", "ifos", "name", "version", "comment"])
    doc.childNodes[0].appendChild(seg_def_table)

    for ifos, name, version, comment, start_time, end_time in distinct_rows:
        seg_def_id                     = seg_def_table.get_next_id()
        segment_definer                = lsctables.SegmentDef()
        segment_definer.process_id     = process_id
        segment_definer.segment_def_id = seg_def_id
        segment_definer.ifos           = ifos.strip()
        segment_definer.name           = name
        segment_definer.version        = version
        segment_definer.comment        = comment

        seg_def_table.append(segment_definer)




def run_defined(doc, process_id, engine, include_segments, times):
    time_clause    = build_time_clause('segment_summary', times)
    segment_clause = ''

    if include_segments:
        segment_clause = ' AND ' + build_segment_clause(include_segments)

    rows = engine.query("""SELECT segment_definer.ifos, segment_definer.name, segment_definer.version, segment_definer.comment,
               segment_summary.start_time, segment_summary.end_time
        FROM segment_definer, segment_summary
        WHERE segment_definer.segment_def_id = segment_summary.segment_def_id
        AND """ + time_clause + segment_clause)

    seg_def_table = lsctables.New(lsctables.SegmentDefTable, columns = ["process_id", "segment_def_id", "ifos", "name", "version", "comment"])
    doc.childNodes[0].appendChild(seg_def_table)

    for ifos, name, version, comment, start_time, end_time in rows:
        seg_def_id                     = seg_def_table.get_next_id()
        segment_definer                = lsctables.SegmentDef()
        segment_definer.process_id     = process_id
        segment_definer.segment_def_id = seg_def_id
        segment_definer.ifos           = ifos.strip()
        segment_definer.name           = name
        segment_definer.version        = version
        segment_definer.comment        = comment

        seg_def_table.append(segment_definer)


#
# =============================================================================
#
#                                 Database/ldbd routines
#
# =============================================================================
#

def ping_server(client):
    print(client.ping())
    return 0


#
# =============================================================================
#
#                                 XML/File routines
#
# =============================================================================
#


def setup_files(dir_name, options, times):
    # Filter out the ones that are outside our time range
    xml_files = []
    for t in times:
        xml_files += segmentdb_utils.get_all_files_in_range(dir_name, t[0], t[1])

    handle, temp_db    = tempfile.mkstemp(suffix='.sqlite')
    os.close(handle)

    target     = dbtables.get_connection_filename(temp_db, None, True, False)
    connection = ligolw_sqlite.setup(target)

    ligolw_sqlite.insert_from_urls(connection, xml_files)

    segmentdb_utils.ensure_segment_table(connection)

    return temp_db, connection
    


#
# =============================================================================
#
#                                     Main
#
# =============================================================================
#

if __name__ == '__main__':
    options, database_location, file_location, times  = parse_command_line()    

    # Ping the database and exit if requested
    if options.ping:
        connection = segmentdb_utils.setup_database(database_location)
        sys.exit( ping_server(connection) )

    start_pad = 0
    end_pad   = 0
    
    if options.start_pad:
        start_pad = int(options.start_pad)

    if options.end_pad:
        end_pad = int(options.end_pad)

    times = [(int(x) - start_pad, int(x) + end_pad) for x in times]

    # set up the response
    doc = ligolw.Document()
    doc.appendChild(ligolw.LIGO_LW())
    process_id = process.register_to_xmldoc(doc, PROGRAM_NAME, options.__dict__, version = git_version.id, cvs_entry_time = __date__).process_id

    temp_files = False

    if database_location:
        connection = segmentdb_utils.setup_database(database_location)
        engine     = query_engine.LdbdQueryEngine(connection)
    else:
        temp_db, connection = setup_files(file_location, options, times)
        engine     = query_engine.SqliteQueryEngine(connection)
        temp_files = True

    
    if options.defined:
        run_defined(doc, process_id, engine, options.include_segments, times)
        utils.write_filename(doc, options.output_file)    

    if options.active:
       run_active(doc, process_id, engine, options.include_segments, times)
       utils.write_filename(doc, options.output_file)

    if options.report:
       run_report(doc, process_id, engine, options.include_segments, times, options.in_segments_only)


    # Clean up
    if temp_files:
        os.remove(temp_db)



